#
# Copyright 2019-Present Sonatype Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""test_ossindex.py audits the call to OSSIndex"""
import unittest
import json
import datetime as DT

from unittest.mock import patch
from pathlib import Path
from datetime import timedelta
from tinydb import TinyDB, Query

from ..ossindex.ossindex import OssIndex
from ..parse.parse import Parse
from ..types.coordinates import Coordinates
from ..types.coordinateresults import CoordinateResults
from ..types.results_decoder import ResultsDecoder
from ..types.vulnerabilities import Vulnerabilities


class TestOssIndex(unittest.TestCase):
  """TestOssIndex audits the call to OSSIndex"""
  def setUp(self):
    self.func = OssIndex(url="http://blahblah",
                         cache_location="/tmp")
    self.parse = Parse()

  def tearDown(self):
    self.func.close_db()
    if Path('/tmp/.ossindex/jake.json').exists():
      Path('/tmp/.ossindex/jake.json').unlink()

  def test_get_headers(self):
    """test_get_headers ensures headers are being retrieved correctly"""
    self.assertEqual(self.func.get_headers(), {
        "Content-type":
        "application/vnd.ossindex.component-report-request.v1+json",
        "User-Agent":
        "jake"})

  def test_get_url(self):
    """test_get_url ensures url is being retrieved correctly"""
    self.assertEqual(self.func.get_url(), "http://blahblah")

  @staticmethod
  def get_fake_purls():
    """get_fake_purls is a helper function that creates a fake Coordinate"""
    fake_purls = Coordinates()
    fake_purls.add_coordinate('thing', '1.1', 'conda')
    fake_purls.add_coordinate('thing', '1.2', 'conda')
    fake_purls.add_coordinate('thing', '1.3', 'conda')
    return fake_purls

  @staticmethod
  def get_fake_actual_purls():
    """get_fake_actual_purls is a helper function that creates a
    fake Coordinate with realistic data"""
    fake_actual_purls = Coordinates()
    fake_actual_purls.add_coordinate('pycrypto', '2.6.1', 'conda')
    return fake_actual_purls

  @patch('jake.ossindex.ossindex.requests.post')
  def test_call_get_dependencies_returns_purls(self, mock_post):
    """test_call_get_dependencies_returns_purls mocks a call to OSSIndex
    and ensures that that calls to OSSIndex return properly formated purls
    as Coordinates or Vulnerabilities type objects"""
    file = Path(__file__).parent / "ossindexresponse.txt"
    with open(file, "r") as stdin:
      mock_result = stdin.read()
      mock_post.return_value.status_code = 200
      mock_post.return_value.text = mock_result
      response = self.func.call_ossindex(self.get_fake_purls())
    self.assertEqual(len(response), 32)
    self.assertEqual(response[0].get_coordinates(),
                     "pkg:conda/pycrypto@2.6.1")
    self.assertEqual(response[18].get_coordinates(),
                     "pkg:conda/python@3.7.3")
    self.assertEqual(isinstance(
        response[18].get_vulnerabilities()[0], Vulnerabilities), True)
    self.assertEqual(response[18].get_vulnerabilities()[
        0].get_id(), "156d71e4-6ed5-4d5f-ae47-7d57be01d387")

  @patch('jake.ossindex.ossindex.requests.post')
  def test_call_ossindex_post_returns_error(self, mock_post):
    """test_call_ossindex_post_returns_error mocks a failed call to OSSIndex
    and ensures the error is handled"""
    mock_post.return_value.status_code = 404
    mock_post.return_value.text = "yadda"
    response = self.func.call_ossindex(self.get_fake_purls())
    self.assertEqual(response, None)

  def test_chunk(self):
    """test_chunk ensures the chunk method is splitting responses with more
    than 128 purl results into 128-purl chunks"""
    file = Path(__file__).parent / "condalistoutput.txt"
    with open(file, "r") as stdin:
      purls = self.parse.get_deps_stdin(stdin)
      actual_result = self.func.chunk(purls)
    self.assertEqual(len(actual_result), 3)
    self.assertEqual(len(actual_result[0]), 128)
    self.assertEqual(actual_result[0][0],
                     "pkg:conda/_ipyw_jlab_nb_ext_conf@0.1.0")
    self.assertEqual(actual_result[1][0], "pkg:conda/mistune@0.8.4")
    self.assertEqual(actual_result[2][0], "pkg:conda/yaml@0.1.7")

  def test_insert_into_cache(self):
    """test_insert_into_cache ensures the results from an OSSIndex call
    are inserted into cache if they do not yet exist in cache"""
    file = Path(__file__).parent / "ossindexresponse.txt"
    with open(file, "r") as stdin:
      response = json.loads(stdin.read(), cls=ResultsDecoder)
      (cached, num_cached) = self.func.maybe_insert_into_cache(response)
    self.assertEqual(num_cached, 32)
    self.assertEqual(cached, True)

  def test_insert_into_cache_does_not_duplicate(self):
    """test_insert_into_cache_does_not_duplicate ensures that maybe_insert_into_cache
    does not insert results into cache if they exist in cache"""
    file = Path(__file__).parent / "ossindexresponse.txt"
    with open(file, "r") as stdin:
      response = json.loads(stdin.read(), cls=ResultsDecoder)
      self.func.maybe_insert_into_cache(response)
      (cached, num_cached) = self.func.maybe_insert_into_cache(response)
    self.assertEqual(num_cached, 0)
    self.assertEqual(cached, False)

  def test_insert_into_cache_expired_ttl(self):
    """test_insert_into_cache_expired_ttl ensures that maybe_insert_into_cache
    inserts results into cache if the time to live is expired"""
    database = TinyDB('/tmp/.ossindex/jake.json')
    coordinate_query = Query()
    response = self.string_to_coordinatesresult(
        """[{"coordinates":"pkg:conda/pycrypto@2.6.1",
        "reference":"https://ossindex.sonatype.org/component/pkg:conda/pycrypto@2.6.1",
        "vulnerabilities":[]}]""")
    self.func.maybe_insert_into_cache(response)
    result_expired = database.search(
        coordinate_query.purl == "pkg:conda/pycrypto@2.6.1")
    time_unwind = DT.datetime.strptime(
      result_expired[0]['ttl'],
      '%Y-%m-%dT%H:%M:%S.%f') - timedelta(hours=13)
    database.update({'ttl': time_unwind.isoformat()},
                    coordinate_query.purl == "pkg:conda/pycrypto@2.6.1")

    next_response = self.string_to_coordinatesresult(
        """[{"coordinates":"pkg:conda/pycrypto@2.6.1",
        "reference":"https://ossindex.sonatype.org/component/pkg:conda/pycrypto@2.6.1",
        "vulnerabilities":[]}]""")
    (cached, num_cached) = self.func.maybe_insert_into_cache(next_response)
    self.assertEqual(cached, True)
    self.assertEqual(num_cached, 1)
    database.close()

  def test_get_purls_from_cache(self):
    """test_get_purls_from_cache ensures that get_purls_and_results_from_cache
    properly retireves purl objs and results from the cache"""
    self.func.maybe_insert_into_cache(self.string_to_coordinatesresult(
        """[{"coordinates":"pkg:conda/pycrypto@2.6.1",
        "reference":"https://ossindex.sonatype.org/component/pkg:conda/pycrypto@2.6.1",
        "vulnerabilities":[{"id":"156d71e4-6ed5-4d5f-ae47-7d57be01d387",
        "title":"[CVE-2019-16056] jake the snake",
        "cvssScore":0.0,"cve":"CVE-2019-16056",
        "reference":"http://www.wrestling.com"}]}]"""))
    (new_purls, results) = self.func.get_purls_and_results_from_cache(
        self.get_fake_actual_purls())
    self.assertEqual(isinstance(results, list), True)
    self.assertEqual(isinstance(results[0], CoordinateResults), True)
    self.assertEqual(results[0].get_coordinates(),
                     "pkg:conda/pycrypto@2.6.1")
    self.assertEqual(results[0].get_reference(
    ), "https://ossindex.sonatype.org/component/pkg:conda/pycrypto@2.6.1")
    self.assertEqual(isinstance(
        results[0].get_vulnerabilities(), list), True)
    self.assertEqual(isinstance(
        results[0].get_vulnerabilities()[0], Vulnerabilities), True)
    self.assertEqual(results[0].get_vulnerabilities()[
        0].get_id(), "156d71e4-6ed5-4d5f-ae47-7d57be01d387")
    self.assertEqual(results[0].get_vulnerabilities()[
        0].get_cve(), "CVE-2019-16056")
    self.assertEqual(len(new_purls.get_coordinates()), 0)
    self.assertEqual(isinstance(new_purls, Coordinates), True)

  def test_get_purls_from_cache_with_cache_miss(self):
    """ This test ensures that a) results can be added to the cache
    and b) if a purl is not in the cache, that purl is still in the new purls
    that are returned """
    self.func.maybe_insert_into_cache(self.string_to_coordinatesresult(
        """[{"coordinates":"pkg:conda/pycrypto@2.6.1",
        "reference":"https://ossindex.sonatype.org/component/pkg:conda/pycrypto@2.6.1",
        "vulnerabilities":[{"id":"156d71e4-6ed5-4d5f-ae47-7d57be01d387",
        "title":"[CVE-2019-16056] jake the snake",
        "cvssScore":0.0,"cve":"CVE-2019-16056",
        "reference":"http://www.wrestling.com"}]}]"""))
    fake_purls = self.get_fake_actual_purls()
    fake_purls.add_coordinate('alabaster', '0.7.12', 'conda')
    (new_purls, results) = self.func.get_purls_and_results_from_cache(
        fake_purls)
    self.assertEqual(len(new_purls.get_coordinates()), 1)
    self.assertEqual(isinstance(new_purls, Coordinates), True)
    self.assertEqual(isinstance(results, list), True)
    self.assertEqual(isinstance(results[0], CoordinateResults), True)
    self.assertEqual(isinstance(
        results[0].get_vulnerabilities()[0],
        Vulnerabilities), True)
    self.assertEqual(results[0].get_vulnerabilities()[
        0].get_id(), "156d71e4-6ed5-4d5f-ae47-7d57be01d387")
    self.assertEqual(results[0].get_vulnerabilities()[
        0].get_cve(), "CVE-2019-16056")
    self.assertEqual(results[0].get_coordinates(),
                     "pkg:conda/pycrypto@2.6.1")
    self.assertEqual(results[0].get_reference(
    ), "https://ossindex.sonatype.org/component/pkg:conda/pycrypto@2.6.1")
    self.assertEqual(new_purls.get_coordinates()
                     [('alabaster', '0.7.12', 'conda')], "pkg:conda/alabaster@0.7.12")

  def test_get_purls_from_cache_with_non_valid_object(self):
    """test_get_purls_from_cache_with_non_valid_object ensures calls to
    get_purls_and_results_from_cache with improper objects returns None for
    new_purls and results"""
    (new_purls, results) = self.func.get_purls_and_results_from_cache(
        "bad data")
    self.assertEqual(new_purls, None)
    self.assertEqual(results, None)

  def test_clean_cache_wipes_database(self):
    """test_clean_cache_wipes_database ensures calls to clean_cache will
    clear out documents from the database"""
    self.func.maybe_insert_into_cache(self.string_to_coordinatesresult(
        """[{"coordinates":"pkg:conda/pycrypto@2.6.1",
        "reference":"https://ossindex.sonatype.org/component/pkg:conda/pycrypto@2.6.1",
        "vulnerabilities":[]}]"""))
    self.assertEqual(self.func.clean_cache(), True)

  @staticmethod
  def string_to_coordinatesresult(string):
    """string_to_coordinatesresult is a helper function that turns a string
    into a CoordinateResult typed object"""
    return json.loads(string, cls=ResultsDecoder)



#
# Copyright 2019-Present Sonatype Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""pip.py gets installed pip dependencies"""
# pylint: disable=protected-access
import logging
import ast
import pkg_resources

from ..types.coordinates import Coordinates

class Pip():
  """pip.py gets installed pip dependencies"""
  def __init__(self, targets=None, requirements_file_path=None):
    self._log = logging.getLogger('jake')
    self._format = "pypi"
    self._working_set = pkg_resources.working_set
    self.requirements_file_path = requirements_file_path
    if targets:
      self._working_set = pkg_resources.WorkingSet(ast.literal_eval(targets))
    self._coords = self.generate_dependencies()

  def generate_dependencies(self, coords=Coordinates()) -> (Coordinates):
    """converts list of pkg_resource.working_set into purl coordinates"""

    if self.requirements_file_path:
      with open(self.requirements_file_path, 'r') as requirements_file:
        for line in requirements_file.readlines():
          name, version = line.strip().split('==')
          coords.add_coordinate(name, version, self._format)
    else:
      for i in iter(self._working_set):
        coords.add_coordinate(i.project_name, i._version, self._format)

    if len(coords.get_coordinates()) == 0:
      return None

    self._coords = coords
    return coords

  def get_dependencies(self) -> (Coordinates):
    """ gets the list of dependencies generatated on init
    Returns:
        Coordinates object -- list of coordinates generated by pkg_resources
    """
    return self._coords



#
# Copyright 2019-Present Sonatype Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""config.py stores OSSIndex credentials"""
import logging
import os

from pathlib import Path

import yaml

class Config():
  """config.py handles getting credentials for OSSIndex or setting them"""
  def __init__(self, config_folder='.ossindex', save_location=''):
    self._log = logging.getLogger('jake')

    self._config_name = '.oss-index-config'
    self._old_config_name = '.jake-config'

    self._username = ""
    self._password = ""

    if save_location != '':
      self._save_location = save_location
    else:
      self._save_location = os.path.join(str(Path.home()), config_folder)

    self.__migrate_config_if_at_jake_location()

  def set_password(self, password):
    """set password for OSSIndex request"""
    self._password = password

  def set_username(self, username):
    """set username for OSSIndex request"""
    self._username = username

  def get_config_from_std_in(self):
    """requests user to input their username and password from stdin"""
    username = input("Please enter your username for your OSS Index account: ")
    password = input("Please enter your API Key for OSS Index: ")

    self.set_username(username)
    self.set_password(password)

    result = self.save_config_to_file(
        {"Username": self._username,
         "Token": self._password},
        self._config_name)

    return result

  def save_config_to_file(self, fields, config_name):
    """save stdin to save_location/config_name"""
    try:
      os.makedirs(os.path.join(self._save_location), exist_ok=True)
      with open(os.path.join(self._save_location, config_name), "w+") as file:
        yaml.dump(fields, file, default_flow_style=False)
        return True
    except FileNotFoundError as exception:
      self._log.error("Uh oh, an error happened: %s", str(exception))
      return False

  def get_config_from_file(self, config_name):
    """get credentials from save_location/config_name"""
    with open(os.path.join(self._save_location, config_name)) as file:
      doc = yaml.full_load(file)
      results = {}
      for item, doc in doc.items():
        results[item] = doc

    return results

  def check_if_config_exists(self, config_location='.oss-index-config'):
    """check to see if config exists at save_location/config_location"""
    config_location = Path(os.path.join(self._save_location, config_location))
    return config_location.exists()

  def __migrate_config_if_at_jake_location(self):
    """check to see if save_location/.jake-config exists, if so migrate to
    .oss-index-config"""
    if self.check_if_config_exists(self._old_config_name):
      self._log.debug("Migrating OSS Index config")
      os.rename(os.path.join(self._save_location, self._old_config_name),
                os.path.join(self._save_location, self._config_name))
    else:
      self._log.debug("No need to migrate config")



#
# Copyright 2019-Present Sonatype Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

""" test_audit.py , for all your testing of audit py needs """
import unittest
import json

from pathlib import Path

from ..audit.audit import Audit
from ..types.results_decoder import ResultsDecoder

class TestAudit(unittest.TestCase):
  """ TestAudit is responsible for testing the Audit class """
  def setUp(self):
    self.func = Audit()

  def test_call_audit_results_prints_output(self):
    """ test_call_audit_results_prints_output ensures that when called with
    a valid result, audit_results returns the number of vulnerabilities found """
    filename = Path(__file__).parent / "ossindexresponse.txt"
    with open(filename, "r") as stdin:
      response = json.loads(
          stdin.read(),
          cls=ResultsDecoder)
    self.assertEqual(self.func.audit_results(response),
                     self.expected_results())

  @staticmethod
  def expected_results():
    """ Weeee, I'm helping! """
    return 3



#
# Copyright 2019-Present Sonatype Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""vulnerabilities.py creates a Vulnerabilities type object"""
# pylint: disable=C0103
# pylint: disable=W0622

class Vulnerabilities():
  """creates a Vulnerabilities type object"""
  def __init__(self):
    self.id = ""
    self.title = ""
    self.description = ""
    self.cvssScore = ""
    self.cvssVector = None
    self.cve = ""
    self.reference = ""

  def set_id(self, id):
    """sets vuln_id for Vulnerabilities object"""
    self.id = id

  def get_id(self):
    """gets vuln_id for Vulnerabilities object"""
    return self.id

  def set_title(self, title):
    """sets title for Vulnerabilities object"""
    self.title = title

  def get_title(self):
    """gets title for Vulnerabilities object"""
    return self.title

  def set_description(self, description):
    """sets description for Vulnerabilities object"""
    self.description = description

  def get_description(self):
    """gets description for Vulnerabilities object"""
    return self.description

  def set_cvss_score(self, cvssScore):
    """sets cvss_score for Vulnerabilities object"""
    self.cvssScore = cvssScore

  def get_cvss_score(self):
    """gets cvss_score for Vulnerabilities object"""
    return self.cvssScore

  def set_cvss_vector(self, cvssVector):
    """sets cvss_vector Vulnerabilities object if it exists"""
    if cvssVector is not None:
      self.cvssVector = cvssVector

  def get_cvss_vector(self):
    """gets cvss_vector for Vulnerabilities object"""
    return self.cvssVector

  def set_cve(self, cve):
    """sets cve for Vulnerabilities object"""
    self.cve = cve

  def get_cve(self):
    """gets cve for Vulnerabilities object"""
    return self.cve

  def set_reference(self, reference):
    """sets reference for Vulnerabilities object"""
    self.reference = reference

  def get_reference(self):
    """gets reference for Vulnerabilities object"""
    return self.reference



#
# Copyright 2019-Present Sonatype Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""test_config.py audits the Config class"""
import unittest

from ..config.config import Config


class TestConfig(unittest.TestCase):
  """testConfig audits the Config class"""
  def setUp(self):
    self.func = Config(save_location='/tmp/')

  def test_config_object_saves_config_file(self):
    """test_config_object_saves_config_file ensures config objs are being written to file"""
    result = self.func.save_config_to_file(
        {"Username": "test@me.com", "Token": "password"},
        ".oss-index-config")
    self.assertEqual(result, True)

  def test_get_config_from_file(self):
    """test_config_object_saves_config_file ensures config objs are being written to file"""
    self.func.save_config_to_file(
        {"Username": "test@me.com", "Token": "password"},
        ".oss-index-config")

    results = self.func.get_config_from_file(".oss-index-config")
    self.assertEqual(results["Username"], "test@me.com")
    self.assertEqual(results["Token"], "password")

  def test_return_true_when_config_exists(self):
    """test_config_object_saves_config_file ensures config objs are being written to file"""
    self.func.save_config_to_file(
        {"Username": "test@me.com", "Token": "password"},
        ".oss-index-config")
    self.assertEqual(self.func.check_if_config_exists(), True)



#
# Copyright 2019-Present Sonatype Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""test_coordinateresults.py audits the CoordinateResult type"""
import unittest
import ast

from ..types.coordinateresults import CoordinateResults


class TestResultsDecoder(unittest.TestCase):
  """TestResultsDecoder ensures CoordinateResults are probably decoded to JSON"""
  def test_to_json_on_coordinateresults_returns_proper_json(self):
    """ensures that the to_json method for the CoordinateResults type returns
     proper JSON"""
    undertest = CoordinateResults()
    undertest.set_coordinates("pkg:conda/thing@1.0.0")
    undertest.set_reference("http://www.wrestling.com")
    undertest.set_vulnerabilities(
        '[{"id":"156d71e4-6ed5-4d5f-ae47-7d57be01d387",'
        '"title":"[CVE-2019-16056]'
        ' jake the snake","cvssScore":0.0,"cve":"CVE-2019-16056"'
        ',"reference":"http://www.wrestling.com"}]')

    result = undertest.to_json()
    dictionary = ast.literal_eval(result)

    self.assertEqual(isinstance(result, str), True)
    self.assertEqual(dictionary['coordinates'], "pkg:conda/thing@1.0.0")
    self.assertEqual(dictionary['reference'], "http://www.wrestling.com")
    self.assertEqual(dictionary['vulnerabilities'],
                     '[{"id":"156d71e4-6ed5-4d5f-ae47-7d57be01d387",'
                     '"title":"[CVE-2019-16056] jake the snake",'
                     '"cvssScore":0.0,"cve":"CVE-2019-16056",'
                     '"reference":"http://www.wrestling.com"}]')



#
# Copyright 2019-Present Sonatype Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""test_main.py test stuff in __main__.py"""
import unittest

from yaspin import yaspin

from jake.__main__ import _show_summary
from jake.iq.iq import IQ


class TestMain(unittest.TestCase):
  """TestMain tests stuff in __main__"""

  def setUp(self):
    iq_args = {}
    self.func = IQ(args=iq_args)

  def test_show_summary(self):
    """test_show_summary verifies exit code is correct for a given IQ Policy Action"""
    spinner = yaspin(text="Loading", color="magenta")

    # pylint: disable=W0212
    self.func._report_url = 'myRelativeReportURL'

    self.func._policy_action = None
    self.assertEqual(3, _show_summary(self.func, spinner))

    self.func._policy_action = "Failure"
    self.assertEqual(1, _show_summary(self.func, spinner))

    self.func._policy_action = "Warning"
    self.assertEqual(0, _show_summary(self.func, spinner))

    self.func._policy_action = "None"
    self.assertEqual(0, _show_summary(self.func, spinner))



#
# Copyright 2019-Present Sonatype Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""jake entry point"""
# pylint: disable=too-many-arguments
# pylint: disable=invalid-name
# pylint: disable=unnecessary-pass

import sys
import io
import logging
from os import _exit, path, mkdir, makedirs
from errno import EACCES, EPERM
from pathlib import Path

import click
from termcolor import cprint
from pyfiglet import figlet_format
from colorama import init, Fore
from yaspin import yaspin

from .ossindex.ossindex import OssIndex
from .iq.iq import IQ
from .cyclonedx.generator import CycloneDxSbomGenerator
from .parse.parse import Parse
from .pip.pip import Pip
from .audit.audit import Audit
from .config.config import Config
from .config.iq_config import IQConfig
from ._version import __version__

# strip colors on redirected output
init(strip=not sys.stdout.isatty())
# save the initial stdout to toggle back to when turned off
_og_stdout = sys.stdout

def __print_version(ctx, flag: bool):
  if not flag:
    return
  print(__package__, 'v' +  __version__)
  ctx.exit()

def __clear_cache(ctx, flag: bool):
  if not flag:
    return
  ossi = OssIndex()
  if ossi.clean_cache():
    print('Cache Cleared')
  ctx.exit()

def __check_stdin(flag: bool):
  if flag and sys.stdin.isatty():
    print('No stdin detected, run \'conda list | jake ...\' with the \'-c\' flag.')
    _exit(0)
  else:
    return

# params that propagate through subcommands
__shared_options = [
    click.option(
        '-vv', '--verbose',
        is_flag=True,
        default=False,
        help='Set log level to verbose'),
    click.option(
        '-q', '--quiet',
        is_flag=True,
        default=False,
        help='Suppress cosmetic and informational output'),
    click.option(
        '-c', '--conda',
        default=False,
        is_flag=True,
        help='Resolve conda dependencies from std_in'),
    click.option(
        '-t', '--targets',
        default=None,
        help="""Specify external site-packages to evaluate.\n
                "`python -c "import site; print(site.getsitepackages())"`"\n
                Passing the above into -t targets site packages for the current shell/venv
             """),
    click.option(
        '-r', '--requirements',
        default=None,
        help='Path of pip requirements file')
]

# decorators be parsed inside out which click handles, but no decorators on the shared options
def __add_options(options):
  def _add_options(func):
    for option in reversed(options):
      func = option(func)
    return func
  return _add_options

@click.group(help='Jake: Put your python deps in a chokehold.')
# options that will take priority over other program execution and exit
@click.option(
    '-v', '--version',
    is_flag=True,
    callback=__print_version,
    expose_value=False,
    is_eager=True,
    help='Print version and exit')
@click.option(
    '--clear',
    is_flag=True,
    callback=__clear_cache,
    expose_value=False,
    is_eager=True,
    help='Clear the OSS Index cache and exit')

# entry point, the above options get executed first as callbacks
def main():
  """ defining the root cli command as main so that running 'jake'
      in the command line will use this as the entry point
      also prints the banner with every invokation

  Arguments:
      version -- jake flag that will print version and exit
      verbose -- get full runtime output from debug logger
      quiet -- supress the banner TODO: non vulnerable outputs as well
  """

@main.command()
@click.argument(
    'conf',
    type=click.Choice(['iq', 'ossi']))
def config(conf):
  """
  Allows a user to set Nexus IQ or OSS Index config params

  Arguments:
      type -- cli input restricted by click to 'iq' and 'ossi'
  """
  cli_config = IQConfig() if conf == 'iq' else Config()

  # exits 0 if config was set, with non-zero from os if it failed
  result = cli_config.get_config_from_std_in()
  if result is False:
    _exit(1)
  else:
    _exit(0)

@main.command()
@__add_options(__shared_options)
@click.option(
    '-o', '--output',
    help='Specify a file name and/or directory to save the CycloneDx sbom.')
def sbom(verbose, quiet, conda, targets, requirements, output):
  """
  Generates a purl only bom (no vulns) and outputs it to std_out
  by default or to a file path with the -o flag

  Does not make any connection to IQ or OSSI

  Arguments:
    output -- file name, relative or absolute path (w/ file name)
  """
  # suppress all output except the sbom xml unless verbose
  if not verbose:
    quiet = __toggle_stdout(on=False)
  __banner(quiet)
  __setup_logger(verbose)
  __check_stdin(conda)

  sbom_xml = __sbom_control_flow(conda, targets, requirements).decode('utf-8')

  # toggle stdout back on and either print to console or to file passed by -o flag
  __toggle_stdout(on=True)
  if not output:
    print(sbom_xml)
  else:
    if path.split(output)[0]:
      try:
        makedirs(path.dirname(output), exist_ok=True)
      except (IOError, OSError) as e:
        if e.errno == EPERM or e.errno == EACCES:
          print("PermissionError({0}) -- {1} for creating directory \'{2}\'".format(
              e.errno,
              e.strerror,
              path.dirname(output)))
          _exit(e.errno)
        else:
          print('Unknown system error: ', sys.exc_info()[0])
          _exit(e.errno)
    with open(output, 'w') as bom_file:
      print(sbom_xml, file=bom_file)
  _exit(0)

# ddt (ossi) subcommand
@main.command()
@__add_options(__shared_options)
@click.option(
    '-o', '--output-format',
    help='Specify ddt output format(json or hr=human-readable)')
def ddt(verbose, quiet, conda, targets, requirements, output_format='hr'):
  """SPECIAL MOVE\n
  Allows you to perform scans backed by Sonatype's OSS Index

  Example usage:\n
      Python scan: jake ddt\n
      Conda scan: conda list | jake ddt -c\n
  """
  __setup_logger(verbose)
  __check_stdin(conda)

  if output_format == 'json':
    __toggle_stdout(on=False)

  __banner(quiet)

  with yaspin(text="Loading", color="yellow") as spinner:
    spinner.text = "Collecting Dependencies"
    coords = (
      Parse().get_deps_stdin(sys.stdin)
      if conda
      else Pip(targets, requirements).get_dependencies()
    )
    spinner.ok("🐍 ")

  with yaspin(text="Loading", color="yellow") as spinner:
    spinner.text = "Querying OSS Index"

    response = OssIndex().call_ossindex(coords)

    if response is None:
      spinner.fail("💥 ")
      click.echo(
          "Something went horribly wrong, there is no response from OSS Index.\n" +
          "please rerun with --verbose to see what happened")
      _exit(1)
    spinner.ok("🐍 ")

  with yaspin(text="Loading", color="yellow") as spinner:
    spinner.text = "Auditing results from OSS Index"
    audit = Audit(quiet)
    spinner.ok("🐍 ")
    __toggle_stdout(on=True)
    code = audit.audit_results(response, output_format=output_format)
    _exit(code)

@main.command()
@__add_options(__shared_options)
@click.option(
    '-i', '--insecure',
    default=False,
    is_flag=True,
    help='Allow jake to communicate with insecure endpoints')
@click.option(
    '-a', '--application',
    help='Supply an IQ Server Public Application ID',
    required=True)
@click.option(
    '-s', '--stage',
    default='develop',
    type=click.Choice(['develop', 'build', 'stage-release', 'release']),
    help='Specify a stage')
@click.option(
    '-u', '--user',
    help='Set username for Sonatype IQ')
@click.option(
    '-p', '--password',
    help='Set password or token for associated user')
@click.option(
    '-h', '--host',
    help='Specify an endpoint for Sonatype IQ')
def iq(verbose: bool, quiet: bool, conda: bool, targets: str, requirements: str,
       insecure: bool, application, stage, user, password, host):
  """EXTRA SPECIAL MOVE\n
  Allows you to perform scans backed by Sonatype's Nexus IQ Server

  Example usage:\n
      Python scan: jake iq -a <AppId>\n
      Conda scan: conda list | jake iq -a <AppId> -c\n

  Will pull values for other params from config unless overwritten here\n

      To set the IQ config: jake config iq\n
  """
  __banner(quiet)
  __setup_logger(verbose)
  __check_stdin(conda)
  bom = __sbom_control_flow(conda, targets, requirements)

  iq_args = {}
  iq_args['application'] = application
  iq_args['stage'] = stage
  iq_args['user'] = user
  iq_args['password'] = password
  iq_args['host'] = host
  iq_args['conda'] = conda
  iq_args['insecure'] = insecure

  __iq_control_flow(iq_args, bom)

def __setup_logger(verbose: bool):
  logger = logging.getLogger('jake')
  logger.setLevel(logging.DEBUG)

  home = str(Path.home())
  if not path.exists(path.join(home, '.ossindex')):
    mkdir(path.join(home, '.ossindex'))

  filepath = path.join(home, '.ossindex', 'jake.combined.log')
  formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')

  fh = logging.FileHandler(filepath)

  logger.addHandler(fh)

  ch = logging.StreamHandler()

  if verbose:
    ch.setLevel(logging.DEBUG)
  else:
    ch.setLevel(logging.ERROR)

  fh.setFormatter(formatter)
  ch.setFormatter(formatter)

  logger.addHandler(ch)

  if verbose:
    print("log file location: "+filepath)

def __iq_control_flow(args: dict, bom_str: bytes):
  with yaspin(text="Loading", color="magenta") as spinner:
    spinner.text = "Submitting to Sonatype IQ..."
    try:
      iq_requests = IQ(args)
    except (ValueError) as e:
      print(e)
      _exit(1)
    status_url = iq_requests.submit_sbom(bom_str)
    spinner.ok("🐍 ")

  with yaspin(text="Loading", color="magenta") as spinner:
    spinner.text = "Reticulating splines..."
    iq_requests.poll_report(status_url)

    _exit(_show_summary(iq_requests, spinner))


def _show_summary(iq_requests, spinner):
  if iq_requests.get_policy_action() == 'Failure':
    spinner.fail("💥 ")
    __toggle_stdout(on=True)
    print(Fore.RED +
          "Snakes on the plane! There are policy failures from Sonatype IQ.")
    print(Fore.RED + "Your IQ Server Report is available here: {}"
          .format(iq_requests.get_absolute_report_url()))
    exit_status_code = 1
  elif iq_requests.get_policy_action() == 'Warning':
    spinner.ok("🧨 ")
    __toggle_stdout(on=True)
    print(Fore.YELLOW +
          "Something slithers around your ankle! There are policy warnings from Sonatype IQ.")
    print(Fore.YELLOW + "Your IQ Server Report is available here: {}"
          .format(iq_requests.get_absolute_report_url()))
    exit_status_code = 0
  elif iq_requests.get_policy_action() == 'None':
    spinner.ok("🐍 ")
    __toggle_stdout(on=True)
    print(Fore.GREEN +
          "Smooth slithering there bud! No policy failures from Sonatype IQ.")
    print(Fore.GREEN + "Your IQ Server Report is available here: {}"
          .format(iq_requests.get_absolute_report_url()))
    exit_status_code = 0
  else:
    spinner.fail("🐍 ")
    __toggle_stdout(on=True)
    print(Fore.RED +
          "Received an error response from IQ Server, please check logs. policy action: {}"
          .format(iq_requests.get_policy_action()))
    exit_status_code = 3

  # reset console colors, etc
  print(Fore.RESET)
  return exit_status_code


def __sbom_control_flow(conda: bool, target: str, requirements_file_path: str) -> (bytes):
  """
  Gets the purls depending on the format and generates the sbom

  Arguments:
      conda -- whether to get conda deps from stdin

  Returns:
      bytestring of the sbom
  """
  with yaspin(text="Loading", color="yellow") as spinner:
    spinner.text = "Collecting Dependencies from System..."
    coords = (
      Parse().get_deps_stdin(sys.stdin)
      if conda
      else Pip(target, requirements_file_path).get_dependencies()
    )
    spinner.ok("🐍 ")
    spinner.text = "Parsing Coordinates..."
    purls = coords.get_purls()
    spinner.ok("🐍 ")

  with yaspin(text="Loading", color="magenta") as spinner:
    spinner.text = "Generating CycloneDx BOM..."
    sbom_gen = CycloneDxSbomGenerator()
    sbom_xml = sbom_gen.purl_sbom(purls)
    sbom_byte_str = sbom_gen.sbom_to_string(sbom_xml)
    spinner.ok("🐍 ")

  return sbom_byte_str

def __banner(quiet: bool):
  """ Prints the banner, most of the user facing commands start with this """
  if quiet:
    # don't turn off entire stdout, otherwise other output (errors) are also suppressed
    # __toggle_stdout(on=False)
    return
  top_font = 'isometric4' # another option: 'isometric1'
  bot_font = 'invita'
  top_text = 'Jake'
  bot_text = ' ..the snake..'
  cprint(figlet_format(top_text, font=top_font), 'green', attrs=[])
  cprint(figlet_format(bot_text, font=bot_font), 'blue', attrs=['dark'])
  click.echo("Jake version: v{}".format(__version__))
  click.echo('Put your python deps in a chokehold.')

def __toggle_stdout(on=True) -> (bool):
  """ Turns console output on and off and returns the state, turns it on by default  """
  sys.stdout = _og_stdout if on else io.StringIO()
  return not on  # return type to set the quiet argument (opposite)

if __name__ == '__main__':
  main()



#
# Copyright 2019-Present Sonatype Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""test_iq_config.py verifies the IQConfig class"""
import io

import sys
import unittest

from ..config.iq_config import IQConfig

class TestIQConfig(unittest.TestCase):
  """testIQConfig verifies the IQConfig class"""
  def setUp(self):
    self.func = IQConfig(save_location='/tmp/')

  def tearDown(self):
    sys.stdin = sys.__stdin__

  def test_get_config_from_std_in(self):
    """test_get_config_from_std_in verifies the IQConfig class"""
    testinput = """test@me.com
password
http://localhost:8070/"""
    sys.stdin = io.StringIO(testinput)
    result = self.func.get_config_from_std_in()
    self.assertEqual(result, True)
    # reread stored config
    results = self.func.get_config_from_file(".iq-server-config")
    self.assertEqual(results["Username"], "test@me.com")
    self.assertEqual(results["Token"], "password")
    self.assertEqual(results["Server"], "http://localhost:8070")



#
# Copyright 2019-Present Sonatype Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""test_pip.py audits the Pip class"""
import unittest

from ..pip.pip import Pip
from ..types.coordinates import Coordinates

class TestPip(unittest.TestCase):
  """TestPip audits the Pip class"""
  def setUp(self):
    self.func = Pip()

  def test_call_get_dependencies_returns_purls(self):
    """test_call_get_dependencies_returns_purls ensures that calls to
    Pip.get_dependencies() returns a Coordinates object"""
    actual = self.func.get_dependencies()

    self.assertIsNotNone(actual)
    self.assertEqual(actual.__class__, Coordinates)



#
# Copyright 2019-Present Sonatype Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""iq_config.py stores IQ credentials"""
from ..config.config import Config

class IQConfig(Config):
  """does IQ specific config setting and retrieval"""
  def __init__(self, save_location=''):
    super().__init__('.iqserver', save_location)
    self._iq_server_config_name = '.iq-server-config'
    self._iq_server_location = ''

  def get_config_from_std_in(self):
    """requests user to input their username and password from stdin"""
    username = input("Please enter your username for your IQ Server account: ")
    password = input("Please enter your user token for IQ Server: ")
    iq_server_location = input("Please provide the location of your IQ Server: ")
    self._iq_server_location = iq_server_location.rstrip("/")
    self.set_username(username)
    self.set_password(password)

    result = self.save_config_to_file(
        {"Username": self._username,
         "Token": self._password,
         "Server": self._iq_server_location},
        self._iq_server_config_name)

    return result



#
# Copyright 2019-Present Sonatype Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""test_results_decoder.py audits the ResultsDecoder class"""
import unittest
import pathlib
import json

from ..types.results_decoder import ResultsDecoder
from ..types.coordinateresults import CoordinateResults


class TestResultsDecoder(unittest.TestCase):
  """TestResultsDecoder audits the ResultsDecoder class"""
  def test_results_decoder_can_transform_ossindex_response(self):
    """test_results_decoder_can_transform_ossindex_response ensures calls
    to ResultsDecorder will return a CoordinateResults typed object"""
    file = pathlib.Path(__file__).parent / "ossindexresponse.txt"
    with open(file, "r") as stdin:
      result = json.loads(stdin.read(), cls=ResultsDecoder)
    self.assertEqual(len(result), 32)
    self.assertEqual(isinstance(result, list), True)
    self.assertEqual(isinstance(result[0], CoordinateResults), True)
    self.assertEqual(result[0].get_coordinates(),
                     "pkg:conda/pycrypto@2.6.1")
    self.assertEqual(result[0].get_reference(
    ), "https://ossindex.sonatype.org/component/pkg:conda/pycrypto@2.6.1")
    self.assertEqual(isinstance(
        result[0].get_vulnerabilities(),
        list), True)
    self.assertEqual(len(result[0].get_vulnerabilities()), 0)



#
# Copyright 2019-Present Sonatype Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""parse.py parses dependencies and converts them to purls"""
import logging
from shutil import which

from ..types.coordinates import Coordinates

class Parse():
  """parse.py parses dependencies and converts them to purls"""
  def __init__(self):
    self._log = logging.getLogger('jake')
    self._format = "conda"

  def get_dependencies(self, run_command_list):
    """checks if conda exists and then gets a list of conda dependencies from stdout"""
    if self.check_if_conda_exists():
      return self.really_get_conda_dependencies(run_command_list)
    return None

  def get_deps_stdin(self, stdin):
    """gets depdencies from stdin"""
    return self.parse_conda_dependencies_into_purls(stdin)

  def check_if_conda_exists(self):
    """checks to see if user installed conda"""
    self._log.debug(which("python"))
    conda_exists = which("conda")

    self._log.debug(conda_exists)

    if conda_exists is not None:
      return True

    return False

  def really_get_conda_dependencies(self, run_command_list):
    """gets a list of installed conda dependencies"""
    results = self.run_conda_list_command(run_command_list)

    length = len(results)

    if results[length-1] == 0:
      return self.parse_conda_dependencies_into_purls(results)

    return None

  @classmethod
  def run_conda_list_command(cls, run_command_list):
    """checks stdout to see if user installed conda"""
    return run_command_list

  def parse_conda_dependencies_into_purls(self, stdin):
    """converts list of dependencies from stdin into purl coordinates"""
    coords = Coordinates()
    self._log.debug("Starting to parse results")
    for line in stdin:
      if "#" in line:
        self._log.debug("Skipping line")
      else:
        line_array = line.split()
        if len(line_array) != 0:
          coords.add_coordinate(line_array[0], line_array[1], self._format)

    if len(coords.get_coordinates()) == 0:
      return None

    return coords



#
# Copyright 2019-Present Sonatype Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""coordinates.py creates a Coordinates type object"""
import json

class Coordinates():
  """
  An object that contains a dict with name/version/source keys and purl values
  It also generates the purl as a string from the coordinate values passed in
  Allows for joins with other Coordinate dicts, adding coordinates, and retrieving
  as a dict or list of purl strings
  """

  def __init__(self):
    self._coordinates = dict()
    # extension needed for IQ ingestion of pypi components
    self._pypi_extension = "?extension=tar.gz"

  def add_coordinate(self, name: str, version: str, source: str):
    """
    adds a coordinate to the object dict based on coordinate values
    """
    purl = self.parse_to_purl(name, version, source)
    self._coordinates[(name, version, source)] = purl

  def get_coordinates(self) -> (dict):
    """gets the coordinates set"""
    return self._coordinates

  def get_purls(self) -> (list):
    """
    gets a list of purls from the objects dict values

    Returns:
        list -- purls as strings
    """
    return list(self._coordinates.values())

  def get_coordinates_as_json(self) -> (str):
    """turns the coordinates array to JSON"""
    coordinates = {}
    coordinates['coordinates'] = self._coordinates
    return json.dumps(coordinates)

  def parse_to_purl(self, name: str, version: str, source: str) -> (str):
    """
    parses a single name, version, source set into a purl

    Returns:
        string -- purl
    """

    template = "pkg:{}/{}@{}"
    purl = template.format(source, name, version)

    # adds the pypi extension for IQ (does not affect oss index results)
    if source == "pypi":
      purl += self._pypi_extension
    return purl

  def join_coords(self, coords: dict):
    """
    Takes in a dict generated from another Coordinates object and joins them,
    then sets the current objects dict to the result

    Returns:
        Coordinates object -- updated object with the joined dict
    """

    self._coordinates.update(coords)
    return self



#
# Copyright 2019-Present Sonatype Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""generator.py will craft a CycloneDX 1.1 SBOM"""
import pathlib
import logging

from lxml import etree

XMLNS = "http://cyclonedx.org/schema/bom/1.1"
XMLNSV = "http://cyclonedx.org/schema/ext/vulnerability/1.0"

NSMAP = {"v" : XMLNSV}

class CycloneDx11Generator():
  """CycloneDx11Generator is responsible for taking identifiers
  and vulnerabilities and turning them into a CycloneDX 1.1 SBOM"""
  def __init__(self):
    self._log = logging.getLogger('jake')
    self.__xml = []

  def create_xml_from_purls(self, coords: list) -> (etree.Element):
    """parses a list of purls to generate an xml w/ only component nodes

    Arguments:
        coords -- list of purls (strings)

    Returns:
        xml -- list of etree.Element objects, component only nodes with no vulns
    """
    self.__create_root()
    components = etree.Element('components')
    for component in coords:
      node = etree.Element('component', {"type": "library", "bom-ref": component})
      nombre, versace = self.__get_name_version_from_purl(component)
      name = etree.SubElement(node, 'name')
      name.text = nombre
      version = etree.SubElement(node, 'version')
      version.text = versace
      purl = etree.SubElement(node, 'purl')
      purl.text = component
      components.append(node)
    self.__xml.append(components)
    return self.__xml

  def create_xml_from_oss_index_results(self, results: list) -> (etree.Element):
    """Takes the CoordinateResults list and creates an sbom in XML form"""
    self.__create_root()
    self.__create_component_nodes(results)
    return self.__xml

  @staticmethod
  def validate_xml_vulnerabilities(xml_vulnerabilities):
    """Validates the given xml against the xsd for vulnerability"""
    file = pathlib.Path(__file__).parent / "vuln.xsd"
    with open(file, "r") as stdin:
      xml_schema_d = etree.parse(stdin)
      xml_schema = etree.XMLSchema(xml_schema_d)
      return xml_schema.assertValid(xml_vulnerabilities)

  def validate_xml(self, xml=None):
    """Takes the XML generated and validates it against the xsd
    for the vulnerability"""
    file = pathlib.Path(__file__).parent / "vuln.xsd"
    with open(file, "r") as stdin:
      xml_schema_d = etree.parse(stdin)
      xml_schema = etree.XMLSchema(xml_schema_d)
      self._log.debug(etree.tostring(self.__xml))
      return xml_schema.assertValid(self.__xml) if xml is None else xml_schema.assertValid(xml)

  def __create_root(self):
    self.__xml = etree.Element('bom', {"xmlns": XMLNS, "version": "1"}, nsmap=NSMAP)

  def __create_component_nodes(self, component_list: list) -> (list):
    components = etree.Element('components')
    for component in component_list:
      node = etree.Element('component', {"type": "library", "bom-ref": component.get_coordinates()})
      nombre, versace = self.__get_name_version_from_purl(component.get_coordinates())
      name = etree.SubElement(node, 'name')
      name.text = nombre
      version = etree.SubElement(node, 'version')
      version.text = versace
      purl = etree.SubElement(node, 'purl')
      purl.text = component.get_coordinates()
      components.append(node)
      if len(component.get_vulnerabilities()) > 0:
        vulnerabilities = etree.Element("{%s}vulnerabilities" % XMLNSV, nsmap=NSMAP)
        self.__create_vulnerability_node(
            component.get_vulnerabilities(),
            component.get_coordinates(),
            vulnerabilities,
            node)
        self.validate_xml_vulnerabilities(vulnerabilities)
    self.__xml.append(components)

  @staticmethod
  def __get_name_version_from_purl(purl):
    split_list = purl.split("/")
    second_split = split_list[1].split("@")
    version_split = second_split[1].split("?")
    return (second_split[0], version_split[0])

  @staticmethod
  def __create_vulnerability_node(vulnerability_list, purl, vulnerabilities, node):
    for vuln in vulnerability_list:
      vulnerability = etree.Element("{%s}vulnerability" % XMLNSV, {"ref": purl})
      _id = etree.SubElement(vulnerability, "{%s}id" % XMLNSV)
      if vuln.get_cve() is None:
        _id.text = "ossindex"
      else:
        _id.text = vuln.get_cve()
      source = etree.Element("{%s}source" % XMLNSV, {"name": "ossindex"})
      url = etree.SubElement(source, "{%s}url" % XMLNSV)
      url.text = vuln.reference
      vulnerability.append(source)
      ratings = etree.Element("{%s}ratings" % XMLNSV)
      rating = etree.SubElement(ratings, "{%s}rating" % XMLNSV)
      score = etree.SubElement(rating, "{%s}score" % XMLNSV)
      base = etree.SubElement(score, "{%s}base" % XMLNSV)
      base.text = str(vuln.get_cvss_score())
      vector = etree.SubElement(rating, "{%s}vector" % XMLNSV)
      vector.text = vuln.get_cvss_vector()
      vulnerability.append(ratings)
      description = etree.SubElement(vulnerability, "{%s}description" % XMLNSV)
      description.text = vuln.description
      vulnerabilities.append(vulnerability)
    node.append(vulnerabilities)



#
# Copyright 2019-Present Sonatype Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""test_parse.py adits the Parse class"""
import unittest
import pathlib

from ..parse.parse import Parse

class TestParse(unittest.TestCase):
  """TestParse audits the Parse class"""
  def setUp(self):
    self.func = Parse()

  def test_call_get_dependencies_returns_purls(self):
    """test_call_get_dependencies_returns_purls ensures that calls to
    Parse.get_coordinates() returns a list of purls"""
    file = pathlib.Path(__file__).parent / "condalistoutput.txt"
    with open(file, "r") as stdin:
      actual = self.func.get_deps_stdin(stdin)
      output = actual.get_coordinates()
    self.assertEqual(len(output), 262)
    self.assertEqual(output[('_ipyw_jlab_nb_ext_conf', '0.1.0', 'conda')],
                     "pkg:conda/_ipyw_jlab_nb_ext_conf@0.1.0")
    self.assertEqual(output[('mkl_fft', '1.0.12', 'conda')],
                     "pkg:conda/mkl_fft@1.0.12")
    self.assertEqual(output[('zstd', '1.3.7', 'conda')],
                     "pkg:conda/zstd@1.3.7")



#
# Copyright 2019-Present Sonatype Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

""" audit.py for all your audit py needs """
# pylint: disable=no-else-return
import json
import logging

from typing import List
from textwrap import wrap
from colorama import Fore
from terminaltables import DoubleTable

from ..types.coordinateresults import CoordinateResults
from ..types.vulnerabilities import Vulnerabilities

class Audit:
  """ Audit does the business, it prints results from OSS Index to the standard out """

  def __init__(self, quiet=False):
    self._log = logging.getLogger("jake")
    self._quiet = quiet

  def audit_results(self, results: List[CoordinateResults], output_format='hr'):
    """
    audit_results is the ingest point for the results from OSS Index,
    and handles control flow
    """
    self._log.debug("Results received, %s total results", len(results))

    if output_format == 'json':
      return self.print_results_json(results)

    total_vulns = 0
    pkg_num = 0
    good = [x for x in results if len(x.get_vulnerabilities()) == 0]
    bad = [x for x in results if len(x.get_vulnerabilities()) > 0]

    if (len(good) != 0 and not self._quiet):
      print()
      print("Non-Vulnerable Dependencies")
      print()
      for coordinate in good:
        pkg_num += 1
        total_vulns += self.print_result(coordinate, pkg_num, len(results))

    if len(bad) != 0:
      print()
      print("Vulnerable Dependencies")
      print()

      for coordinate in bad:
        pkg_num += 1
        total_vulns += self.print_result(coordinate, pkg_num, len(results))

    table_data = [
        ["Audited Dependencies", len(results)],
        ["Vulnerablities Found", total_vulns],
    ]

    table_instance = DoubleTable(table_data, "Summary")

    print()

    print(table_instance.table)

    return total_vulns

  @classmethod
  def print_results_json(cls, results: List[CoordinateResults]):
    """
    print_json takes a list of coordinate results,
    and prints a json array of all vulnerabilities
    """
    vulnerabilities = []
    for result in results:
      for vulnerability in result.get_vulnerabilities():
        vulnerabilities.append({
          "package": result.get_coordinates()[4:],
          "title": vulnerability.get_title(),
          "cvss_score": vulnerability.get_cvss_score(),
          "cve": vulnerability.get_cve(),
          "description": vulnerability.get_description(),
        })
    print(json.dumps(vulnerabilities))
    return sum(len(x.get_vulnerabilities()) for x in results)

  def print_result(self, coordinate: CoordinateResults, number, length):
    """
    print_results takes a coordinate, the index of the coordinate in a list,
    and the length, and handles printing it in different formats if a
    vulnerability exists
    """
    if len(coordinate.get_vulnerabilities()) == 0:
      if not self._quiet:
        self.do_print(
            f"[{number}/{length}] - {coordinate.get_coordinates()}",
            0
        )
      return len(coordinate.get_vulnerabilities())
    self.do_print(
        f"[{number}/{length}] - {coordinate.get_coordinates()} [VULNERABLE]",
        coordinate.get_max_cvss_score(),
    )
    self.do_print(
        f"{len(coordinate.get_vulnerabilities())} known vulnerabilities for this version",
        coordinate.get_max_cvss_score(),
    )
    for vulnerability in coordinate.get_vulnerabilities():
      self.print_vulnerability(vulnerability)
    return len(coordinate.get_vulnerabilities())

  @classmethod
  def print_vulnerability(cls, vulnerability: Vulnerabilities):
    """
    print_vulnerability takes a vulnerability, and well, it prints it
    """
    cvss_score = vulnerability.get_cvss_score()
    table_data = [
        ["ID", vulnerability.get_id()],
        ["Title", vulnerability.get_title()],
        ["Description", '\n'.join(wrap(vulnerability.get_description(), 100))],
        ["CVSS Score", f"{vulnerability.get_cvss_score()} - {cls.get_cvss_severity(cvss_score)}"],
        ]
    if vulnerability.get_cvss_vector():
      table_data.append(
          ["CVSS Vector", vulnerability.get_cvss_vector()]
          )

    table_data.extend(
        [
            ["CVE", vulnerability.get_cve()],
            ["Reference", vulnerability.get_reference()]
        ]
        )
    table_instance = DoubleTable(table_data)
    table_instance.inner_heading_row_border = False
    table_instance.inner_row_border = True
    cls.do_print(table_instance.table, cvss_score)

    print("----------------------------------------------------")

  @classmethod
  def do_print(cls, text, cvss_score):
    """
    do_print takes text, and a cvss_score and prints it in a different text depending on
    the score
    """
    if cvss_score == 0:
      print(Fore.GREEN + text + Fore.RESET)
    elif 0 < cvss_score < 4:
      print(Fore.CYAN + text + Fore.RESET)
    elif 4 <= cvss_score < 7:
      print(Fore.LIGHTYELLOW_EX + text + Fore.RESET)
    elif 7 <= cvss_score < 9:
      print(Fore.YELLOW + text + Fore.RESET)
    else:
      print(Fore.RED + text + Fore.RESET)

  @classmethod
  def get_cvss_severity(cls, cvss_score):
    """
    get_cvss_severity takes a cvss_score and returns a human readable severity for it
    """
    if cvss_score == 0:
      return "None"
    elif 0 < cvss_score < 4:
      return "Low"
    elif 4 <= cvss_score < 7:
      return "Medium"
    elif 7 <= cvss_score < 9:
      return "High"
    else:
      return "Critical"



#
# Copyright 2019-Present Sonatype Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""iq.py handles requests to IQ Server"""
# pylint: disable=too-many-instance-attributes
import json
import logging
import urllib

from json import JSONDecodeError

import requests
import polling

from ..config.iq_config import IQConfig

DEFAULT_HEADERS = {
    'User-Agent': 'jake'}

LOG = logging.getLogger('jake')

class IQ():
  """IQ handles requests to IQ Server"""
  def __init__(self, args):
    self._iq_url = args.get('host')
    self._user = args.get('user')
    self._password = args.get('password')
    self._public_application_id = args.get('application')
    self._stage = args.get('stage')
    self._insecure = args.get('insecure')
    self._headers = DEFAULT_HEADERS
    self._report_url = ''
    self._policy_action = None
    self._request = requests.Session()
    self._internal_id = ''

    if self._insecure:
      self._request.verify = False

    config = IQConfig()
    if config.check_if_config_exists('.iq-server-config') is False:
      LOG.debug("No IQ server config supplied, using defaults or taking from command-line.")
      if self._user is None:
        self._user = 'admin'
      if self._password is None:
        self._password = 'admin123'
      if self._iq_url is None:
        self._iq_url = 'http://localhost:8070'
    else:
      LOG.debug("Found iq server config.  Using those unless overwritten by command line params.")
      results = config.get_config_from_file(".iq-server-config")
      if self._user is None:
        self._user = results['Username']
      if self._password is None:
        self._password = results['Token']
      if self._iq_url is None:
        self._iq_url = results['Server']

    self._request.auth = requests.auth.HTTPBasicAuth(self._user, self._password)

  def get_policy_action(self):
    """gets policy action from IQ Server result"""
    return self._policy_action

  def get_report_url(self):
    """gets report url from IQ Server result"""
    return self._report_url

  def get_absolute_report_url(self):
    """get full (non-relative) url to IQ Server report"""
    # urljoin() will ignore the first parameter (base) if the second parameter (url) is absolute
    return urllib.parse.urljoin(self._iq_url, self._report_url)

  def get_public_application_id(self) -> (str):
    """gets public application id to use for IQ Server request"""
    return self._public_application_id

  def get_internal_id(self) -> (str):
    """gets internal application id from IQ Server using the public
    application id"""
    response = self._request.get(
        '{0}/api/v2/applications?publicId={1}'.format(
            self._iq_url,
            self._public_application_id),
        headers=self._headers)
    if response.ok:
      res = json.loads(response.text)
      if not res['applications']:
        raise ValueError(
            "\nThe public application id \'"
            + self._public_application_id
            + "\' does not exist or is not accessible by the user.")
      LOG.debug(res['applications'][0]['id'])
      return res['applications'][0]['id']
    raise ValueError('\n' + response.text + '\nSet your config with \'jake config iq\'')

  def submit_sbom(self, sbom: str) -> (str):
    """submits sbom (in str form) to IQ server, valid sbom should get
    202 response. On valid response, sets status url for later polling"""
    self._internal_id = self.get_internal_id()

    LOG.debug(sbom)
    headers = self._headers
    headers['Content-Type'] = 'application/xml'
    response = self._request.post(
        '{0}/api/v2/scan/applications/{1}/sources/jake?stageId={2}'.format(
            self._iq_url,
            self._internal_id,
            self._stage),
        data=sbom,
        headers=headers)
    if response.ok:
      res = json.loads(response.text)
      LOG.debug(res['statusUrl'])
      return res['statusUrl']
    raise ValueError(response.text)

  def poll_report(self, status_url: str):
    """polls status url once a second until it gets a 200 response
    , and times out after one minute"""
    polling.poll(
        lambda: self._request.get(
            '{0}/{1}'.format(self._iq_url, status_url)).text,
        check_success=self.__handle_response,
        step=1,
        timeout=60)

  def __handle_response(self, response: str) -> (bool):
    try:
      res = json.loads(response)
      LOG.debug(res)
      self._policy_action = res['policyAction']
      LOG.debug("Policy Action parsed")
      self._report_url = res['reportHtmlUrl']
      return True
    except JSONDecodeError as json_decode_error:
      LOG.debug(json_decode_error.msg)
      return False



#
# Copyright 2019-Present Sonatype Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""coordinateresults.py creates a CoordinateResults type object"""
import json

class CoordinateResults():
  """CoordinateResults creates a CoordinateResults type object"""
  def __init__(self):
    self.coordinates = ""
    self.reference = ""
    self.vulnerabilities = []

  def set_coordinates(self, coordinate):
    """sets coordinates for CoordinateResults obj"""
    self.coordinates = coordinate

  def get_coordinates(self):
    """gets coordinates for CoordinateResults obj"""
    return self.coordinates

  def set_reference(self, reference):
    """sets reference for CoordinateResults obj"""
    self.reference = reference

  def get_reference(self):
    """get reference for CoordinateResults obj"""
    return self.reference

  def set_vulnerabilities(self, vulnerabilities):
    """sets vulnerabilities for CoordinateResults obj"""
    self.vulnerabilities = vulnerabilities

  def get_vulnerabilities(self):
    """sets vulnerabilities for CoordinateResults obj"""
    return self.vulnerabilities

  def get_max_cvss_score(self):
    """gets the max cvss_score for vulnerabilities list"""
    # if not self.vulnerabilities:
    #   return None
    return max(vulnerability.get_cvss_score() for vulnerability in self.vulnerabilities)

  def to_json(self):
    """converts CoordinateResults obj to JSON"""
    return json.dumps(self.__dict__, default=lambda o: o.__dict__)



#
# Copyright 2019-Present Sonatype Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""test_sbom_generator.py audits the cyclonedx/1.1/CycloneDx11Generator class"""
import unittest
import pathlib
import json

from lxml import etree

from ..types.coordinateresults import CoordinateResults
from ..types.results_decoder import ResultsDecoder
from ..cyclonedx.generator import CycloneDxSbomGenerator

class TestSbomGenerator(unittest.TestCase):
  """TestSbomGenerator audits the cyclonedx/CycloneDxSbomGenerator class"""
  def setUp(self):
    self.func = CycloneDxSbomGenerator()

  def test_invalid_bom_version(self):
    """test_invalid_bom_version verifies using a non-implemented bom version fails"""
    self.assertRaises(NotImplementedError, CycloneDxSbomGenerator, "1.0")

  def test_can_create_valid_root_element(self):
    """test_can_create_valid_root_element tests if an sbom can be created
    using the cyclonedx/CycloneDxSbomGenerator class"""
    file = pathlib.Path(__file__).parent / "ossindexvulnerablesnipresponse.txt"
    with open(file, "r") as stdin:
      result = json.loads(stdin.read(), cls=ResultsDecoder)
    results = self.func.create_and_return_sbom(result)
    # Assert that it has a <bom>
    self.assertIsNotNone(results)
    self.assertEqual(etree.iselement(results), True)
    self.assertEqual(results.tag, 'bom')
    # Assert that it has a <components>
    self.assertIs(results.__len__(), 1, results.__len__())
    item = results.__getitem__(0)
    self.assertIsNotNone(item)
    self.assertEqual(item.tag, "components")
    self.assertIs(item.__len__(), 15)
    component = item.__getitem__(0)
    self.assertIsNotNone(component)
    self.assertEqual(component.tag, "component")
    vulnerable_component = item.__getitem__(14)
    self.assertEqual(vulnerable_component.__getitem__(0).text, "ncurses")
    self.assertEqual(vulnerable_component.__getitem__(1).text, "6.1")
    self.assertEqual(vulnerable_component.__getitem__(2).text, "pkg:conda/ncurses@6.1")
    self.assertIsNotNone(vulnerable_component)
    self.assertEqual(vulnerable_component.__len__(), 4)
    vulnerabilities = vulnerable_component.__getitem__(3)
    self.assertEqual(vulnerabilities.__len__(), 5)

  def test__get_name_version_from_purl(self):
    """test__get_name_version_from_purl tests if a parameter suffix is removed from the
    sbom version field"""
    coord_result = CoordinateResults()
    coord_result.set_coordinates("pkg:pypi/yaspin@0.16.0?extension=tar.gz")
    coord_result_normal = CoordinateResults()
    coord_result_normal.set_coordinates("pkg:pypi/normalpurl@0.17.0")
    coord_results = [coord_result, coord_result_normal]
    sbom = self.func.create_and_return_sbom(coord_results)
    # Assert that it has a <bom>
    self.assertIsNotNone(sbom)
    self.assertEqual(etree.iselement(sbom), True)
    self.assertEqual(sbom.tag, 'bom')
    # Assert that it has a <components>
    self.assertIs(sbom.__len__(), 1, sbom.__len__())
    item = sbom.__getitem__(0)
    self.assertIsNotNone(item)
    self.assertEqual(item.tag, "components")
    self.assertIs(item.__len__(), 2)

    component = item.__getitem__(0)
    self.assertIsNotNone(component)
    self.assertEqual(component.tag, "component")
    self.assertEqual(component.__getitem__(0).text, "yaspin")
    self.assertEqual(component.__getitem__(1).text, "0.16.0")
    self.assertEqual(component.__getitem__(2).text, "pkg:pypi/yaspin@0.16.0?extension=tar.gz")
    self.assertIsNotNone(component)

    component_normal = item.__getitem__(1)
    self.assertIsNotNone(component_normal)
    self.assertEqual(component_normal.tag, "component")
    self.assertEqual(component_normal.__getitem__(0).text, "normalpurl")
    self.assertEqual(component_normal.__getitem__(1).text, "0.17.0")
    self.assertEqual(component_normal.__getitem__(2).text, "pkg:pypi/normalpurl@0.17.0")
    self.assertIsNotNone(component_normal)



#
# Copyright 2019-Present Sonatype Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from setuptools import setup, find_packages

from jake._version import __version__

with open("README.md", "r", encoding="utf8") as fh:
  LONG_DESCRIPTION = fh.read()

with open('requirements.txt') as requirements:
  REQUIRED = requirements.read().splitlines()

setup(
    name='jake',
    version=__version__,
    url="https://github.com/sonatype-nexus-community/jake",
    author="Sonatype Community",
    author_email="community-group@sonatype.com",
    description="An OSS Index integration to check for vulnerabilities in your Conda environments",
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    license="Apache-2.0",
    packages=find_packages(exclude=["jake/tests/*", "jake/tests/"]),
    python_requires='>=3.6',
    package_data={
        'jake':
            ['cyclonedx/v1_1/*.xsd']
    },
    install_requires=REQUIRED,
    entry_points={
        'console_scripts':
            ['jake=jake.__main__:main']
    }
)



#
# Copyright 2019-Present Sonatype Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""ossindex.py makes a request to OSSIndex"""
import logging
import json
import datetime as DT

from typing import List
from datetime import datetime, timedelta
from pathlib import Path
from tinydb import TinyDB, Query

import requests

from ..types.coordinates import Coordinates
from ..types.results_decoder import ResultsDecoder
from ..types.coordinateresults import CoordinateResults
from ..config.config import Config

DEFAULT_HEADERS = {
    'Content-type': 'application/vnd.ossindex.component-report-request.v1+json',
    'User-Agent': 'jake'}

class OssIndex():
  """ossindex.py makes a request to OSSIndex"""
  def __init__(self,
               url='https://ossindex.sonatype.org/api/v3/component-report',
               cache_location=''):
    self._url = url
    self._headers = DEFAULT_HEADERS
    self._log = logging.getLogger('jake')
    self._maxcoords = 128
    if cache_location == '':
      home = str(Path.home())
      dir_oss = home + "/.ossindex/"
    else:
      dir_oss = cache_location + "/.ossindex/"
    if not Path(dir_oss).exists():
      Path(dir_oss).mkdir(parents=True, exist_ok=True)
    self._db = TinyDB(dir_oss + "jake.json")

  def get_url(self):
    """gets url to use for OSSIndex request"""
    return self._url

  def get_headers(self):
    """gets headers to use for OSSIndex request"""
    return self._headers

  def chunk(self, coords: Coordinates):
    """chunks up purls array into 128-purl subarrays"""
    chunks = []
    divided = []
    length = len(coords.get_coordinates())
    num_chunks = length // self._maxcoords
    if length % self._maxcoords > 0:
      num_chunks += 1
    start_index = 0
    end_index = self._maxcoords
    for i in range(0, num_chunks):
      if i == (num_chunks - 1):
        divided = coords.get_purls()[start_index:length]
      else:
        divided = coords.get_purls()[start_index:end_index]
        start_index = end_index
        end_index += end_index
      chunks.append(divided)
    return chunks

  def call_ossindex(self, coords: Coordinates) -> (list):
    """makes a request to OSSIndex"""
    self._log.debug("Purls received, total purls before chunk: %s",
                    len(coords.get_coordinates()))

    (coords, results) = self.get_purls_and_results_from_cache(coords)

    self._log.debug("Purls checked against cache, total purls remaining to "
                    "call OSS Index: %s",
                    len(coords.get_coordinates()))

    chunk_purls = self.chunk(coords)
    for purls_chunk in chunk_purls:
      data = {}
      data["coordinates"] = purls_chunk
      config_file = Config()
      if config_file.check_if_config_exists() is False:
        response = requests.post(self.get_url(), data=json.dumps(
            data), headers=self.get_headers())
      else:
        auth = config_file.get_config_from_file(
            ".oss-index-config")

        response = requests.post(self.get_url(),
                                 data=json.dumps(data),
                                 headers=self.get_headers(),
                                 auth=(auth["Username"], auth["Token"]))
      if response.status_code == 200:
        self._log.debug(response.headers)
        first_results = json.loads(response.text, cls=ResultsDecoder)
      else:
        self._log.debug("Response failed, status: %s",
                        response.status_code)
        self._log.debug("Failure reason if any: %s", response.reason)
        self._log.debug("Failure text if any: %s", response.text)
        return None
      results.extend(first_results)

    (cached, num_cached) = self.maybe_insert_into_cache(results)
    self._log.debug("Cached: <%s> num_cached: <%s>", cached, num_cached)
    return results

  def maybe_insert_into_cache(self, results: List[CoordinateResults]):
    """checks to see if result is in cache and if not, stores it"""
    coordinate_query = Query()
    num_cached = 0
    cached = False
    for coordinate in results:
      mydatetime = datetime.now()
      twelvelater = mydatetime + timedelta(hours=12)
      result = self._db.search(
          coordinate_query.purl == coordinate.get_coordinates())
      if len(result) == 0:
        self._db.insert({'purl': coordinate.get_coordinates(),
                         'response': coordinate.to_json(),
                         'ttl': twelvelater.isoformat()})
        self._log.debug(
            "Coordinate inserted into cache: <%s>",
            coordinate.get_coordinates())
        num_cached += 1
        cached = True
      else:
        timetolive = DT.datetime.strptime(
          result[0]['ttl'],
          '%Y-%m-%dT%H:%M:%S.%f'
          )
        if mydatetime > timetolive:
          self._db.update({'response': coordinate.to_json(),
                           'ttl': twelvelater.isoformat()},
                          doc_ids=[result[0].doc_id])
          self._log.debug(
              "Coordinate: <%s> updated in cache because TTL"
              " expired",
              coordinate.get_coordinates())
          num_cached += 1
          cached = True

    return (cached, num_cached)

  def get_purls_and_results_from_cache(self, purls: Coordinates) -> (Coordinates, list):
    """get cached purls and results from cache"""
    valid = isinstance(purls, Coordinates)
    if not valid:
      return (None, None)
    new_purls = Coordinates()
    results = []
    coordinate_query = Query()
    for coordinate, purl in purls.get_coordinates().items():
      mydatetime = datetime.now()
      result = self._db.search(coordinate_query.purl == purl)
      if len(result) == 0 or DT.datetime.strptime(
        result[0]['ttl'],
        '%Y-%m-%dT%H:%M:%S.%f'
        ) < mydatetime:
        new_purls.add_coordinate(coordinate[0], coordinate[1], coordinate[2])
      else:
        results.append(json.loads(
            result[0]['response'], cls=ResultsDecoder))
    return (new_purls, results)

  def clean_cache(self):
    """removes all documents from the table"""
    self._db.truncate()
    return True

  def close_db(self):
    """closes connection to TinyDB"""
    self._db.close()



#
# Copyright 2019-Present Sonatype Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""generator.py will craft and validate a CycloneDX SBOM"""
# pylint: disable=protected-access
import logging

from lxml import etree

from ..cyclonedx.v1_1.generator import CycloneDx11Generator

class CycloneDxSbomGenerator():
  """CycloneDxGenerator is responsible for taking identifiers
  and vulnerabilities and turning them into a CycloneDX SBOM.
  By default it will generate a 1.1 version of the SBOM, if sbom_version
  is set to a different version (and an accompanying implementation is
  done) it can create it. Currently only 1.1 is implemented, any other value
  for sbom_version will throw a NotImplementedError"""
  def __init__(self, sbom_version="1.1"):
    self._log = logging.getLogger('jake')
    if sbom_version == "1.1":
      self.__generator = CycloneDx11Generator()
    else:
      raise NotImplementedError

  def purl_sbom(self, purls: list) -> (etree.Element):
    """ get sbom from a list of purls

    Arguments:
        purls -- list of purls (strings)

    Returns:
        sbom as a list of lxml.etree.Element nodes
        coordinate data only
    """
    sbom = self.__generator.create_xml_from_purls(purls)
    return sbom

  def create_and_return_sbom(self, results) -> (list):
    """create_and_return_sbom is responsible for taking results in
    CoordinateResults form and turning them into a valid CycloneDX SBOM"""
    sbom = self.__generator.create_xml_from_oss_index_results(results)
    return sbom

  @staticmethod
  def sbom_to_string(sbom: etree.Element) -> (bytes):
    """sbom_to_string is responsible for turning an sbom into a string"""
    return etree.tostring(sbom, encoding="UTF-8")

  def validate_sbom(self, sbom):
    """validate_sbom is responsible for taking an sbom in etree Element
     form and validating it against an internal CycloneDX XSD"""
    valid = self.__generator.validate_xml(sbom)
    if valid:
      return sbom
    raise ValueError("Something is wrong with the OSS Index results")



#
# Copyright 2019-Present Sonatype Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""test_iq.py audits the call to IQ"""
import unittest
import json

from pathlib import Path

import responses

from ..iq.iq import IQ

class TestIQ(unittest.TestCase):
  """TestIQ audits the call to IQ"""
  def setUp(self):
    iq_args = {}
    iq_args['application'] = 'testapp'
    iq_args['stage'] = 'develop'
    iq_args['user'] = 'admin'
    iq_args['password'] = 'admin123'
    iq_args['host'] = 'http://afakeurlthatdoesnotexist.com:8081'
    iq_args['conda'] = False
    iq_args['insecure'] = False

    self.internal_id = '4537e6fe68c24dd5ac83efd97d4fc2f4'
    self.status_id = '9cee2b6366fc4d328edc318eae46b2cb'
    self.third_party_url = '{0}/api/v2/scan/applications/{1}/sources/jake?stageId={2}'.format(
      iq_args['host'],
      self.internal_id,
      iq_args['stage']
    )
    self.status_url = 'api/v2/scan/applications/{0}/status/{1}'.format(
      self.internal_id,
      self.status_id
    )
    self.full_status_url = '{0}/{1}'.format(
      iq_args['host'],
      self.status_url
    )

    self.func = IQ(args=iq_args)

  @responses.activate
  def test_call_get_application_id(self):
    """test_call_get_application_id mocks a call to IQ
    and ensures that that calls to IQ for an application ID return
    an internal ID as expected"""
    file = Path(__file__).parent / "iqapplicationresponse.txt"
    with open(file, "r") as stdin:
      responses.add(responses.GET,
                    'http://afakeurlthatdoesnotexist.com:8081/api/v2/applications?publicId=testapp',
                    body=stdin.read(), status=200)
      response = self.func.get_internal_id()
    self.assertEqual(response, self.internal_id)

  @responses.activate
  def test_call_submit_sbom(self):
    """test_call_submit_sbom mocks a call to IQ
    and ensures that that calls to IQ with an sbom return
    a status URL as expected"""
    file = Path(__file__).parent / "iqapplicationresponse.txt"
    with open(file, "r") as stdin:
      responses.add(responses.GET,
                    'http://afakeurlthatdoesnotexist.com:8081/api/v2/applications?publicId=testapp',
                    body=stdin.read(), status=200)
    file = Path(__file__).parent / "iqstatusurlresponse.txt"
    with open(file, "r") as stdin:
      responses.add(responses.POST,
                    self.third_party_url,
                    body=stdin.read(), status=200)
      response = self.func.submit_sbom("sbom")
    self.assertEqual(response, self.status_url)

  @responses.activate
  def test_call_poll_report_none_policy_action(self):
    """test_call_poll_report_none_policy_action mocks a call to IQ
    and ensures that that calls to IQ asking for a poll
    response on a status URL, return as expected, and sets policy action
    to None"""
    file = Path(__file__).parent / "iqpolicynoneresponse.txt"
    with open(file, "r") as stdin:
      mock_json = json.loads(stdin.read())
      responses.add(responses.GET,
                    self.full_status_url,
                    json=mock_json, status=200)

      self.func.poll_report(self.status_url)
    self.assertEqual(self.func.get_policy_action(), 'None')

  @responses.activate
  def test_call_poll_report_failure_policy_action(self):
    """test_call_poll_report_failure_policy_action mocks a call to IQ
    and ensures that that calls to IQ asking for a poll
    response on a status URL, return as expected, and sets policy action
    to Failure"""
    file = Path(__file__).parent / "iqpolicyfailureresponse.txt"
    with open(file, "r") as stdin:
      mock_json = json.loads(stdin.read())
      responses.add(responses.GET,
                    self.full_status_url,
                    json=mock_json, status=200)

      self.func.poll_report(self.status_url)
    self.assertEqual(self.func.get_policy_action(), 'Failure')

  @responses.activate
  def test_call_poll_report_warning_policy_action(self):
    """test_call_poll_report_warning_policy_action mocks a call to IQ
    and ensures that that calls to IQ asking for a poll
    response on a status URL, return as expected, and sets policy action
    to Warning"""
    file = Path(__file__).parent / "iqpolicywarningresponse.txt"
    with open(file, "r") as stdin:
      mock_json = json.loads(stdin.read())
      responses.add(responses.GET,
                    self.full_status_url,
                    json=mock_json, status=200)

      self.func.poll_report(self.status_url)
    self.assertEqual(self.func.get_policy_action(), 'Warning')

  @responses.activate
  def test_call_poll_report_absolute_result_url(self):
    """test_call_poll_report_relative_result_url mocks a call to IQ
    and ensures an absolute report URL is returned,
    and ensure that same absolute report URL is returned from get_absolute_report_url()"""
    file = Path(__file__).parent / "iqresponse_absolute_report_url.json"
    with open(file, "r") as stdin:
      mock_json = json.loads(stdin.read())
      responses.add(responses.GET,
                    self.full_status_url,
                    json=mock_json, status=200)

      self.func.poll_report(self.status_url)
    # pylint: disable=W0212
    self.assertEqual(self.func._iq_url, 'http://afakeurlthatdoesnotexist.com:8081')
    self.assertEqual(self.func.get_report_url(),
                     'http://localhost:8070/ui/links/application/my-app/report/95c4c14e')
    # ensure we use the IQ provided absolute URL
    self.assertEqual(self.func.get_absolute_report_url(),
                     'http://localhost:8070/ui/links/application/my-app/report/95c4c14e')

  @responses.activate
  def test_call_poll_report_relative_result_url(self):
    """test_call_poll_report_relative_result_url mocks a call to IQ
    and ensures a relative report URL is returned,
    and ensure get_absolute_report_url() returns an absolute URL
    built from the _iq_url and relative report URL"""
    file = Path(__file__).parent / "iqresponse_relative_report_url.json"
    with open(file, "r") as stdin:
      mock_json = json.loads(stdin.read())
      responses.add(responses.GET,
                    self.full_status_url,
                    json=mock_json, status=200)

      self.func.poll_report(self.status_url)
    # pylint: disable=W0212
    self.assertEqual(self.func._iq_url, 'http://afakeurlthatdoesnotexist.com:8081')
    self.assertEqual(self.func.get_report_url(),
                     'ui/links/application/my-app/report/95c4c14e')
    # build the absolute URL from the IQ Server base url
    self.assertEqual(self.func.get_absolute_report_url(),
        'http://afakeurlthatdoesnotexist.com:8081/ui/links/application/my-app/report/95c4c14e')



#
# Copyright 2019-Present Sonatype Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""results_decoder.py takes the JSON results from a call to OSSIndex and
 turns them into CoordinateResults or Vulnerabilities type objects"""
import json

from ..types.coordinateresults import CoordinateResults
from ..types.vulnerabilities import Vulnerabilities


class ResultsDecoder(json.JSONDecoder):
  """ResultsDecoder takes the JSON results from a call to OSSIndex and
 turns them into CoordinateResults or Vulnerabilities type objects"""
  def __init__(self):
    json.JSONDecoder.__init__(self, object_hook=self.dict_to_object)

  @classmethod
  def dict_to_object(cls, dictionary):
    """checks to see if dictionary item has coordinates key then if it does
 converts item into CoordinateResults, if it doesnt converts it to
 Vulnerabilities """
    if dictionary.get('coordinates') is not None:
      item = CoordinateResults()
      item.set_coordinates(dictionary.get("coordinates"))
      item.set_reference(dictionary.get("reference"))
      item.set_vulnerabilities(dictionary.get("vulnerabilities"))

      return item

    vulnerability = Vulnerabilities()
    vulnerability.set_id(dictionary.get("id"))
    vulnerability.set_title(dictionary.get("title"))
    vulnerability.set_description(dictionary.get("description"))
    vulnerability.set_cvss_score(dictionary.get("cvssScore"))
    vulnerability.set_cvss_vector(dictionary.get('cvssVector'))
    vulnerability.set_cve(dictionary.get("cve"))
    vulnerability.set_reference(dictionary.get("reference"))

    return vulnerability
